/****************************************************************************
 File name:  prand.c
 Author:     Chadd Williams
 Date:       2/25/07
 Class:      CS360 Special Topics: Computer Networking
 Assignment: Pthreads Programming Example 
 Purpose:    This projects use the pthread library to display two streams
 							of numbers to the screen. One stream of random even numbers
 							and one stream of random odd numbers.  A single thread should
 							generate all the random numbers.
****************************************************************************/

#include <stdio.h>
#include <pthread.h>
#include <stdlib.h>

#define HALF_SECOND 500000
#define QUART_SECOND 250000

// gEven is the shared data used to pass a random number from the random
// thread to the even number printing thread
int gEven = 0;
 
// gOdd is the shared data used to pass a random number from the random
// thread to the odd number printing threadint gOdd=1;
int gOdd = 1;

// The following variables are used to stop the threads.  Each of these
// variables is shared between the UI thread and one of the other threads.
// They are not protected by a mutex since 1) only the UI thread updates the
// variable, 2) if the variable is read and updated at the same time the 
// worst that will happen is that the other thread will run one extra 
// iteration.
int gRandCont = 1;
int gEvenCont = 1;
int gOddCont = 1;

// These mutexes guard the varialbes gEven and gOdd, respectively
// In this assignment, you want a mutex per shared data item so that threads
// can run as freely as possible.  See a note about efficiency in even_thread 
pthread_mutex_t gEvenM;
pthread_mutex_t gOddM;

/****************************************************************************
 Function:    killEven

 Description: This function stops the even thread by setting gEvenCont 

 Parameters:  None

 Returned:    None
****************************************************************************/
void killEven()
{
	gEvenCont = 0;
}

/****************************************************************************
 Function:    killOdd

 Description: This function stops the odd thread by setting gOddCont 

 Parameters:  None

 Returned:    None
****************************************************************************/
void killOdd()
{
	gOddCont = 0;
}

/****************************************************************************
 Function:    killRand

 Description: This function stops the random number thread by setting 
 								gRandCont 

 Parameters:  None

 Returned:    None
****************************************************************************/
void killRand()
{
	gRandCont = 0;
}

/****************************************************************************
 Function:    even_thread

 Description: This is the main function for the even thread.  This thread
 								checks the gEven value, if it is even it sleeps for 1/2 a 
 								second, increments the value by 2, and prints out the value 
 								(preceeded by a single tab).  The thread then unconditionally
 								sleeps for 1/2 a second.
 				
 							This function depends on gEvenCont to stop the thread.

 Parameters:  None

 Returned:    None
****************************************************************************/
void* even_thread(void* pArg)
{
	
	while( gEvenCont )
	{
		/* This is not the most efficient way to do things.
		 * 
		 * It is, however, a good learning example and gives interesting
		 * results if the sleep()s or lock/unlock are removed.
		 * 
		 * To do things most efficiently, you want to hold a lock for as
		 * small amount of time as possible.  A better solution would be
		 * something like this:
		 * 
		 * int localEven; // declare a value to hold the even number locally
		 * 
		 * pthread_mutex_lock(&gEvenM);
		 * localEven = gEven;
		 * pthread_mutex_unlock(&gEvenM);
		 * 
		 * Use the localEven variable in the if, increment, and printf
		 * statement below.
		 * 
		 * This can also be done with the odd thread
		 */
		pthread_mutex_lock(&gEvenM);
		if( gEven % 2 == 0)
		{
			usleep(HALF_SECOND);
			gEven += 2;
			fprintf(stderr, "\t%d\n", gEven);
		}
		pthread_mutex_unlock(&gEvenM);
		usleep(HALF_SECOND);
	}
	pthread_exit(NULL);
}

/****************************************************************************
 Function:    odd_thread

 Description: This is the main function for the odd thread.  This thread
 								checks the gOdd value, if it is odd it sleeps for 1/2 a 
 								second, increments the value by 2, and prints out the value 
 								(preceeded by exactly two tabs).  The thread then 
 								unconditionally sleeps for 1/2 a second.
 				
 							This function depends on gOddCont to stop the thread.

 Parameters:  None

 Returned:    None
****************************************************************************/
void* odd_thread(void* pArg)
{

	while( gOddCont )
	{
		// see comment on efficiency in even_thread
		pthread_mutex_lock(&gOddM);
		if ( gOdd % 2 == 1)
		{
			usleep(HALF_SECOND);
			gOdd += 2;
			fprintf(stderr, "\t\t%d\n", gOdd);
		}
		pthread_mutex_unlock(&gOddM);
		usleep(HALF_SECOND);
	}
	pthread_exit(NULL);
}

/****************************************************************************
 Function:    rand_thread

 Description: This is the main function for the random number generating 
 								thread.  This thread	generates a random number in the 
 								range 0..1000, inclusive.  The random number is alternately
 								place in memory shared with the even and odd thread.  After
 								generating a random number, the thread sleeps for 1/4 second
 				
 							This function depends on gRandCont to stop the thread.

 Parameters:  A pointer to an int to be used as the random number seed.

 Returned:    None
****************************************************************************/
void* rand_thread(void* pArg)
{
	int randNum;
	// this flag tracks which thread (even or odd) gets the next random #
	int flag = 0;

	srand( *(int*) pArg);

	while(gRandCont)
	{
		// to generate a number in the range 0..1000, inclusive, use
		// 1001 in the modulus operation!!
		randNum = rand() % 1001;
		if( flag == 0 )
		{
			//even
			pthread_mutex_lock(&gEvenM);
			gEven = randNum;
			pthread_mutex_unlock(&gEvenM);
			flag = 1;
		}
		else
		{
			//odd
			pthread_mutex_lock(&gOddM);
			gOdd = randNum;
			pthread_mutex_unlock(&gOddM);
			flag = 0;
		}

		usleep(QUART_SECOND);
	}
	pthread_exit(NULL);
}

/****************************************************************************
 Function:    main

 Description: This is the main function.  This initializes the mutexes,
 								spawns three threads and this function acts as the UI thread.
 								When the user inputs a 'q', the other three threads are 
 								stopped and once they terminate this thread (the main
 								program) exits. 
 
 Parameters:  None

 Returned:    None
****************************************************************************/
int main()
{

	pthread_t evenT;
	pthread_t oddT;
	pthread_t randT;
	int seed;
	char command;

	// initialize the mutexes
	pthread_mutex_init(&gEvenM, NULL);
	pthread_mutex_init(&gOddM, NULL);

	// This is the UI thread. Don't continue until the
	// user enters a random seed
	fprintf(stdout," Please enter the seed> ");
	scanf("%d", &seed);

	// spawn the threads
	pthread_create(&evenT, NULL, even_thread, NULL);
	pthread_create(&oddT, NULL, odd_thread, NULL);
	pthread_create(&randT, NULL, rand_thread, (void*) &seed);
	
	// wait until the user inputs a 'q'
	do{
		scanf("%c", &command);
	}while(command != 'q');

	// kill the threads.
	killEven();
	killOdd();
	killRand();
	
	// wait for the threads to finish
	pthread_join(evenT, NULL);
	pthread_join(oddT, NULL);
	pthread_join(randT, NULL);

	fprintf(stdout,"\nTerminating...\n");
	return 0;
}

/* if you open this file in vim, the following line will set your tab
 * size to 2
 */
 
/* vim: set ts=2: */
